/**
 * Ext GWT 3.0.0-rc2 - Ext for GWT
 * Copyright(c) 2007-2012, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.dnd.core.client;

import java.util.List;

import com.google.gwt.user.client.Element;
import com.sencha.gxt.core.client.util.Format;
import com.sencha.gxt.dnd.core.client.DND.Operation;
import com.sencha.gxt.messages.client.DefaultMessages;
import com.sencha.gxt.widget.core.client.grid.Grid;

/**
 * A <code>DragSource</code> implementation for Grids. Supports both single and
 * multi-selections. The selected model instances will be set as the data to be
 * dragged.
 */
public class GridDragSource<M> extends DragSource {

  public interface GridDragSourceMessages {

    String itemsSelected(int items);
  }
  protected class DefaultGridDragSourceMessages implements GridDragSourceMessages {

    public String itemsSelected(int items) {
      return DefaultMessages.getMessages().listField_itemsSelected(items);
    }

  }

  protected Grid<M> grid;

  private GridDragSourceMessages messages;

  /**
   * Creates a new grid drag source instance.
   * 
   * @param grid the grid
   */
  public GridDragSource(Grid<M> grid) {
    super(grid);
    this.grid = grid;
  }

  /**
   * Returns the sources grid component.
   * 
   * @return the grid
   */
  public Grid<M> getGrid() {
    return grid;
  }

  public GridDragSourceMessages getMessages() {
    if (messages == null) {
      messages = new DefaultGridDragSourceMessages();
    }
    return messages;
  }

  public void setMessages(GridDragSourceMessages messages) {
    this.messages = messages;
  }

  @Override
  protected void onDragDrop(DndDropEvent event) {
    if (event.getOperation() == Operation.MOVE) {
      Object data = event.getData();
      if (data instanceof List) {
        @SuppressWarnings("unchecked")
        List<M> list = (List<M>) data;
        for (M item : list) {
          grid.getStore().remove(item);
        }
      }
    }
  }

  @Override
  protected void onDragStart(DndDragStartEvent event) {
    Element r = grid.getView().findRow(event.getDragStartEvent().getStartElement()).cast();
    if (r == null) {
      event.setCancelled(true);
      return;
    }
    List<M> sel = grid.getSelectionModel().getSelectedItems();
    if (sel.size() > 0) {
      event.setCancelled(false);
      event.setData(sel);

      if (getStatusText() == null) {
        event.getStatusProxy().update(getMessages().itemsSelected(sel.size()));
      } else {
        event.getStatusProxy().update(Format.substitute(getStatusText(), sel.size()));
      }
    }
  }

}
