/**
 * Ext GWT 3.0.0-rc2 - Ext for GWT
 * Copyright(c) 2007-2012, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.theme.base.client.grid;

import com.google.gwt.core.client.GWT;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.NodeList;
import com.google.gwt.resources.client.ClientBundle;
import com.google.gwt.resources.client.CssResource;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.resources.client.ImageResource.ImageOptions;
import com.google.gwt.resources.client.ImageResource.RepeatStyle;
import com.google.gwt.safehtml.shared.SafeHtml;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.safehtml.shared.SafeHtmlUtils;
import com.sencha.gxt.core.client.XTemplates;
import com.sencha.gxt.core.client.dom.DomQuery;
import com.sencha.gxt.core.client.dom.XElement;
import com.sencha.gxt.core.client.resources.StyleInjectorHelper;
import com.sencha.gxt.widget.core.client.grid.GridView.GridAppearance;
import com.sencha.gxt.widget.core.client.grid.GridView.GridStyles;

public class GridBaseAppearance implements GridAppearance {

  public interface GridResources extends ClientBundle {

    @Source("Grid.css")
    GridStyle css();

    @ImageOptions(repeatStyle = RepeatStyle.Vertical)
    ImageResource specialColumn();

    @ImageOptions(repeatStyle = RepeatStyle.Vertical)
    ImageResource specialColumnSelected();

  }

  public interface GridStyle extends CssResource, GridStyles {

    String columnLines();

    String empty();

    String grid();

    String scroller();

    String body();

  }

  public interface GridTemplates extends XTemplates {
    @XTemplate(source = "Grid.html")
    SafeHtml render(GridStyle style);
  }

  private final GridResources resources;
  private final GridStyle style;
  private GridTemplates templates = GWT.create(GridTemplates.class);

  public GridBaseAppearance() {
    this(GWT.<GridResources> create(GridResources.class));
  }

  public GridBaseAppearance(GridResources resources) {
    this.resources = resources;
    this.style = this.resources.css();

    StyleInjectorHelper.ensureInjected(style, true);
  }

  @Override
  public void render(SafeHtmlBuilder sb) {
    sb.append(templates.render(style));
  }

  @Override
  public GridStyles styles() {
    return style;
  }

  @Override
  public Element findRow(Element elem) {
    if (Element.is(elem)) {
      return elem.<XElement> cast().findParentElement("." + style.row(), 15);
    }
    return null;
  }

  @Override
  public NodeList<Element> getRows(XElement parent) {
    return DomQuery.select("." + style.row(), parent);
  }

  @Override
  public Element findCell(Element elem) {
    if (Element.is(elem)) {
      return elem.<XElement> cast().findParentElement("." + style.cell(), 15);
    }
    return null;
  }

  @Override
  public void onRowOver(Element row, boolean over) {
    row.<XElement> cast().setClassName(style.rowOver(), over);
  }

  @Override
  public void onRowHighlight(Element row, boolean highlight) {
    row.<XElement> cast().setClassName(style.rowHighlight(), highlight);
  }

  @Override
  public void onRowSelect(Element row, boolean select) {
    row.<XElement> cast().setClassName(style.rowSelected(), select);
  }

  @Override
  public void onCellSelect(Element cell, boolean select) {
    cell.<XElement> cast().setClassName(style.cellSelected(), select);
  }

  @Override
  public Element getRowBody(Element row) {
    return row.<XElement> cast().selectNode("." + style.rowBody());
  }

  @Override
  public SafeHtml renderEmptyContent(String emptyText) {
    return SafeHtmlUtils.fromTrustedString("<div class=\"" + style.empty() + "\">" + emptyText + "<div>");
  }

}
